@@ -895,25 +895,44 @@ prepare_context (gst_context_part context,
 		 int args,
 		 int temps)
 {
-  REGISTER (1, OOP * mySP);
-  _gst_temporaries = mySP = context->contextStack;
+  REGISTER (1, OOP *stackBase);
+  _gst_temporaries = stackBase = context->contextStack;
   if (args)
     {
-      REGISTER (2, int num);
-      REGISTER (3, OOP * src);
-      num = args;
-      src = &sp[1 - num];
-
-#define UNROLL_OP(n) mySP[n] = src[n]
-#define UNROLL_ADV(n) mySP += n, src += n
-      UNROLL_BY_8 (num);
-#undef UNROLL_OP
-#undef UNROLL_ADV
-
-      mySP += num;
+      REGISTER (2, OOP * src);
+      src = &sp[1 - args];
+      stackBase[0] = src[0];
+      if (args > 1)
+        {
+          stackBase[1] = src[1];
+          if (args > 2)
+            {
+              stackBase[2] = src[2];
+              if (args > 3)
+                memcpy (&stackBase[3], &src[3], (args - 3) * sizeof (OOP));
+            }
+        }
+      stackBase += args;
     }
-  sp = mySP + temps - 1;
-  nil_fill (mySP, temps);
+  if (temps)
+    {
+      REGISTER (2, OOP src);
+      src = _gst_nil_oop;
+      stackBase[0] = src;
+      if (temps > 1)
+        {
+          stackBase[1] = src;
+          if (temps > 2)
+            {
+              int n = 2;
+              do
+                stackBase[n] = src;
+              while UNCOMMON (n++ < temps);
+            }
+        }
+      stackBase += temps;
+    }
+  sp = stackBase - 1;
 }
 
 mst_Boolean
@@ -1015,11 +1034,9 @@ unwind_context (void)
 {
   gst_method_context oldContext, newContext;
   OOP oldContextOOP, newContextOOP;
-  int numLifoContexts;
 
   newContextOOP = _gst_this_context_oop;
   newContext = (gst_method_context) OOP_TO_OBJ (newContextOOP);
-  numLifoContexts = free_lifo_context - lifo_contexts;
 
   do
     {
@@ -1029,21 +1046,18 @@ unwind_context (void)
       /* Descend in the chain...  */
       newContextOOP = oldContext->parentContext;
 
-      if COMMON (numLifoContexts > 0)
-	{
-	  dealloc_stack_context ((gst_context_part) oldContext);
-	  numLifoContexts--;
-	}
+      if COMMON (free_lifo_context > lifo_contexts)
+        dealloc_stack_context ((gst_context_part) oldContext);
 
-      else
-	/* This context cannot be deallocated in a LIFO way.  We must
-	   keep it around so that the blocks it created can reference
-	   arguments and temporaries in it. Method contexts, however,
-	   need to be marked as non-returnable so that attempts to
-	   return from them to an undefined place will lose; doing
-	   that for block contexts too, we skip a test and are also
-	   able to garbage collect more context objects.  */
-	oldContext->parentContext = _gst_nil_oop;
+      /* This context cannot be deallocated in a LIFO way.  We must
+         keep it around so that the blocks it created can reference
+         arguments and temporaries in it. Method contexts, however,
+         need to be marked as non-returnable so that attempts to
+         return from them to an undefined place will lose; doing
+         that for block contexts too, we skip a test and are also
+         able to garbage collect more context objects.  And doing
+         that for _all_ method contexts is more icache-friendly.  */
+      oldContext->parentContext = _gst_nil_oop;
 
       newContext = (gst_method_context) OOP_TO_OBJ (newContextOOP);
     }
@@ -1568,10 +1582,13 @@ mst_Boolean
 _gst_sync_signal (OOP semaphoreOOP, mst_Boolean incr_if_empty)
 {
   gst_semaphore sem;
-  OOP freedOOP;
+  gst_process process;
+  gst_method_context suspendedContext;
+  OOP processOOP;
+  int spOffset;
 
   sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
-  for (;;)
+  do
     {
       /* printf ("signal %O %O\n", semaphoreOOP, sem->firstLink); */
       if (is_empty (semaphoreOOP))
@@ -1581,12 +1598,20 @@ _gst_sync_signal (OOP semaphoreOOP, mst_Boolean incr_if_empty)
 	  return false;
 	}
 
-      freedOOP = remove_first_link (semaphoreOOP);
+      processOOP = remove_first_link (semaphoreOOP);
 
       /* If they terminated this process, well, try another */
-      if (resume_process (freedOOP, false))
-	return true;
     }
+  while (!resume_process (processOOP, false));
+
+  /* Put the semaphore at the stack top as a marker that the
+     wait was not interrupted.  This assumes that _gst_sync_wait
+     is only called from primitives.  */
+  process = (gst_process) OOP_TO_OBJ (processOOP);
+  suspendedContext = (gst_method_context) OOP_TO_OBJ (process->suspendedContext);
+  spOffset = TO_INT (suspendedContext->spOffset);
+  suspendedContext->contextStack[spOffset] = semaphoreOOP;
+  return true;
 }
 
 static void
@@ -1666,7 +1691,12 @@ _gst_sync_wait (OOP semaphoreOOP)
   sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
   if (TO_INT (sem->signals) <= 0)
     {
-      /* have to suspend, move this to the end of the list */
+      /* Have to suspend.  Prepare return value for #wait and move
+         this process to the end of the list.
+
+         Tweaking the stack top means that this function should only
+	 be called from a primitive.  */
+      SET_STACKTOP (_gst_nil_oop);
       add_last_link (semaphoreOOP, get_active_process ());
       if (IS_NIL (ACTIVE_PROCESS_YIELD ()))
         {
